# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
import re
from typing import Dict, List, Tuple

import pytest

from swh.objstorage.backends.in_memory import InMemoryObjStorage
from swh.objstorage.exc import (
    NoBackendsLeftError,
    ObjCorruptedError,
    ReadOnlyObjStorageError,
)
from swh.objstorage.multiplexer import (
    MP_BACKEND_DISABLED_METRICS,
    MP_BACKEND_ENABLED_METRICS,
    MP_COUNTER_METRICS,
    MultiplexerObjStorage,
)
from swh.objstorage.objstorage import DURATION_METRICS, compute_hashes

from .objstorage_testing import ObjStorageTestFixture


def statsd_payloads_having_tags(
    statsd, **tags: str
) -> List[Tuple[str, Dict[str, str]]]:
    ret = []
    for s in statsd.socket.payloads:
        m = re.fullmatch(
            r"^(?P<metric_name>[^:]+):(?P<value>[^|]+)[|](?P<unit>ms|c)[|](#(?P<tags>.*))?$",
            s.decode(),
        )
        if not m:
            continue
        payload_tags = {}
        for tag in m.group("tags").split(","):
            tag, value = tag.split(":", 1)
            payload_tags[tag] = value
        if tags.items() <= payload_tags.items():
            ret.append((m.group("metric_name"), payload_tags))
    return ret


def clear_statsd_payloads(statsd) -> None:
    statsd.socket.payloads.clear()


class TestMultiplexerObjStorage(ObjStorageTestFixture):
    @pytest.fixture
    def swh_objstorage_config(self, tmpdir):
        root1 = os.path.join(tmpdir, "root1")
        root2 = os.path.join(tmpdir, "root2")
        os.mkdir(root1)
        os.mkdir(root2)
        return {
            "cls": "multiplexer",
            "objstorages": [
                {
                    "cls": "read-only",
                    "name": "ro_backend",
                    "storage": {
                        "cls": "pathslicing",
                        "name": "pathslicer_1",
                        "root": root1,
                        "slicing": "0:2/2:4",
                    },
                },
                {
                    "cls": "pathslicing",
                    "name": "rw_backend",
                    "root": root2,
                    "slicing": "0:1/0:5",
                },
            ],
        }

    def test_contains(self):
        content_p, obj_id_p = self.hash_content(b"contains_present")
        content_m, obj_id_m = self.hash_content(b"contains_missing")
        self.storage.add(content_p, obj_id=obj_id_p)
        assert obj_id_p in self.storage
        assert obj_id_m not in self.storage

    @pytest.fixture
    def allow_delete(self):
        for storage in self.storage.storages:
            storage.allow_delete = True

    def test_delete_missing(self, allow_delete):
        super().test_delete_missing()

    def test_delete_missing_composite(self, allow_delete):
        super().test_delete_missing_composite()

    def test_delete_present(self, allow_delete):
        super().test_delete_present()

    def test_delete_present_composite(self, allow_delete):
        super().test_delete_present_composite()

    def test_access_readonly(self):
        # Add a content to the readonly storage
        content, obj_id = self.hash_content(b"content in read-only")
        with pytest.raises(ReadOnlyObjStorageError):
            self.storage.storages[0].add(content, obj_id=obj_id)
        # Try to retrieve it on the main storage
        assert obj_id not in self.storage

    def test_get_statsd_multiplexer(self, statsd):
        content1, obj_id1 = self.compositehash_content(b"add_get_batch_1")
        content2, obj_id2 = self.compositehash_content(b"add_get_batch_2")
        content3, obj_id3 = self.compositehash_content(b"add_get_batch_3")
        self.storage.add(content1, obj_id1)
        self.storage.add(content2, obj_id2)
        expected_payloads = [
            # first call to add()
            (DURATION_METRICS, {"endpoint": "add", "name": "multiplexer"}),
            # second call to add()
            (DURATION_METRICS, {"endpoint": "add", "name": "multiplexer"}),
        ]
        payloads = statsd_payloads_having_tags(statsd, name="multiplexer")
        assert payloads == expected_payloads

        # add a content in the RO backend
        self.storage.storages[0].storage.add(content3, obj_id3)
        clear_statsd_payloads(statsd)

        self.storage.get(obj_id1)
        self.storage.get(obj_id2)
        self.storage.get(obj_id3)

        # check metrics reported by the multiplexer itself
        expected_mp_payloads = [
            # first get()
            (
                MP_COUNTER_METRICS,
                {
                    "backend": "rw_backend",
                    "backend_number": "1",
                    "endpoint": "get",
                    "name": "multiplexer",
                },
            ),
            (DURATION_METRICS, {"endpoint": "get", "name": "multiplexer"}),
            # second get()
            (
                MP_COUNTER_METRICS,
                {
                    "backend": "rw_backend",
                    "backend_number": "1",
                    "endpoint": "get",
                    "name": "multiplexer",
                },
            ),
            (DURATION_METRICS, {"endpoint": "get", "name": "multiplexer"}),
            # third get()
            (
                MP_COUNTER_METRICS,
                {
                    "backend": "ro_backend",
                    "backend_number": "0",
                    "endpoint": "get",
                    "name": "multiplexer",
                },
            ),
            (DURATION_METRICS, {"endpoint": "get", "name": "multiplexer"}),
        ]
        mp_payloads = statsd_payloads_having_tags(statsd, name="multiplexer")
        assert mp_payloads == expected_mp_payloads

        # check metrics reported by the ro backend (aka pathslicer_1)
        expected_ro_payloads = [
            # first get()
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
            (
                f"{DURATION_METRICS}_error_count",
                {
                    "endpoint": "get",
                    "error_type": "ObjNotFoundError",
                    "name": "pathslicer_1",
                },
            ),
            # second get()
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
            (
                f"{DURATION_METRICS}_error_count",
                {
                    "endpoint": "get",
                    "error_type": "ObjNotFoundError",
                    "name": "pathslicer_1",
                },
            ),
            # third get()
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
            (DURATION_METRICS, {"endpoint": "get", "name": "pathslicer_1"}),
        ]
        ro_payloads = statsd_payloads_having_tags(statsd, name="pathslicer_1")
        assert ro_payloads == expected_ro_payloads

        # check metrics reported by the rw backend (aka rw_backend)
        expected_rw_payloads = [
            # first get()
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "rw_backend"}),
            (DURATION_METRICS, {"endpoint": "get", "name": "rw_backend"}),
            # second get()
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "rw_backend"}),
            (DURATION_METRICS, {"endpoint": "get", "name": "rw_backend"}),
        ]
        rw_payloads = statsd_payloads_having_tags(statsd, name="rw_backend")
        assert rw_payloads == expected_rw_payloads

        # clear the statsd messages
        statsd.socket.payloads.clear()
        # and check metrics for the __contains__ method
        assert obj_id1 in self.storage
        assert obj_id2 in self.storage
        assert obj_id3 in self.storage

        # for now, there is no counter (MP_COUNTER_METRICS) on any other method
        # than get, so we don't have any simple stat on who answered the
        # request...
        expected_mp_payloads = [
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "multiplexer"}),
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "multiplexer"}),
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "multiplexer"}),
        ]
        mp_payloads = statsd_payloads_having_tags(statsd, name="multiplexer")
        assert mp_payloads == expected_mp_payloads

        expected_ro_payloads = [
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "pathslicer_1"}),
        ]
        ro_payloads = statsd_payloads_having_tags(statsd, name="pathslicer_1")
        assert ro_payloads == expected_ro_payloads

        expected_rw_payloads = [
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "rw_backend"}),
            (DURATION_METRICS, {"endpoint": "__contains__", "name": "rw_backend"}),
        ]
        rw_payloads = statsd_payloads_having_tags(statsd, name="rw_backend")
        assert rw_payloads == expected_rw_payloads


def test_multiplexer_corruption_fallback(mocker, caplog):
    content_p = b"contains_present"
    obj_id_p = compute_hashes(content_p)

    class CorruptedInMemoryObjStorage(InMemoryObjStorage):
        name = "corrupted_objstorage"

        def get(self, obj_id):
            raise ObjCorruptedError("Always corrupted", obj_id)

    corrupt_storage = CorruptedInMemoryObjStorage()
    corrupt_get = mocker.spy(corrupt_storage, "get")

    ok_storage = InMemoryObjStorage()
    ok_get = mocker.spy(ok_storage, "get")

    multiplexer = MultiplexerObjStorage(objstorages=[corrupt_storage, ok_storage])
    multiplexer.add(content_p, obj_id=obj_id_p)

    assert obj_id_p in corrupt_storage
    assert obj_id_p in ok_storage

    with caplog.at_level(logging.WARNING, "swh.core.statsd"), caplog.at_level(
        logging.WARNING, "swh.objstorage.multiplexer"
    ):
        assert multiplexer.get(obj_id_p) == content_p

    corrupt_get.assert_called_once_with(obj_id_p)
    ok_get.assert_called_once_with(obj_id_p)

    assert len(caplog.records) == 1
    assert (
        "was reported as corrupted by backend 'corrupted_objstorage'"
        in caplog.records[0].message
    )


def test_multiplexer_transient_error_fallback(mocker, caplog, statsd):
    content_p = b"contains_present"
    obj_id_p = compute_hashes(content_p)

    class TimeoutInMemoryObjStorage(InMemoryObjStorage):
        name = "timeout-in-memory"

        def get(self, obj_id):
            raise TimeoutError("Always timeout", obj_id)

    timeout_storage = TimeoutInMemoryObjStorage(name="always-timeout")
    timeout_get = mocker.spy(timeout_storage, "get")

    ok_storage = InMemoryObjStorage()
    ok_get = mocker.spy(ok_storage, "get")

    multiplexer = MultiplexerObjStorage(
        objstorages=[timeout_storage, ok_storage],
        read_exception_cooldown=2,
        name="my-multiplexer",
    )
    multiplexer.add(content_p, obj_id=obj_id_p)

    assert obj_id_p in timeout_storage
    assert obj_id_p in ok_storage

    with caplog.at_level(
        logging.WARNING, "swh.objstorage.multiplexer.multiplexer_objstorage"
    ):
        assert multiplexer.get(obj_id_p) == content_p

    timeout_get.assert_called_once_with(obj_id_p)
    ok_get.assert_called_once_with(obj_id_p)

    assert len(caplog.records) == 1
    assert "always-timeout" in caplog.records[0].message
    assert "transient" in caplog.records[0].message
    for algo, hash in obj_id_p.items():
        assert f"{algo}:{hash.hex()}" in caplog.records[0].message

    assert 0 in multiplexer.reset_timers

    assert (
        MP_BACKEND_DISABLED_METRICS,
        {
            "backend": "always-timeout",
            "backend_number": "0",
            "endpoint": "get",
            "name": "my-multiplexer",
        },
    ) in statsd_payloads_having_tags(statsd, name="my-multiplexer")

    clear_statsd_payloads(statsd)

    timeout_get.reset_mock()
    assert multiplexer.get(obj_id_p) == content_p

    timeout_get.assert_not_called()

    assert MP_BACKEND_DISABLED_METRICS not in (
        metric
        for metric, tags in statsd_payloads_having_tags(statsd, name="my-multiplexer")
    )
    clear_statsd_payloads(statsd)

    multiplexer.enable_backend(name="always-timeout", i=0)

    assert (
        MP_BACKEND_ENABLED_METRICS,
        {
            "backend": "always-timeout",
            "backend_number": "0",
            "name": "my-multiplexer",
        },
    ) in statsd_payloads_having_tags(statsd, name="my-multiplexer")
    clear_statsd_payloads(statsd)

    assert multiplexer.get(obj_id_p) == content_p
    timeout_get.assert_called_once_with(obj_id_p)

    assert (
        MP_BACKEND_DISABLED_METRICS,
        {
            "backend": "always-timeout",
            "backend_number": "0",
            "endpoint": "get",
            "name": "my-multiplexer",
        },
    ) in statsd_payloads_having_tags(statsd, name="my-multiplexer")


def test_multiplexer_transient_error_nobackendsleft(mocker, caplog):
    content_p = b"contains_present"
    obj_id_p = compute_hashes(content_p)

    class TimeoutInMemoryObjStorage(InMemoryObjStorage):
        def get(self, obj_id):
            raise TimeoutError("Always timeout", obj_id)

    multiplexer = MultiplexerObjStorage(objstorages=[TimeoutInMemoryObjStorage()])
    multiplexer.add(content_p, obj_id=obj_id_p)

    with pytest.raises(NoBackendsLeftError):
        multiplexer.get(obj_id=obj_id_p)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
from collections import Counter
import concurrent.futures
import datetime
import logging
from multiprocessing import current_process
import os
import random
import sys
import time
from typing import Any, Dict, Optional, Set, Union

import psycopg
import psycopg_pool
from typing_extensions import Literal

from swh.objstorage.backends.winery.objstorage import (
    WineryObjStorage,
    WineryReader,
    WineryWriter,
    rw_shard_cleaner,
    shard_packer,
)
from swh.objstorage.backends.winery.roshard import Pool
from swh.objstorage.backends.winery.sharedbase import ShardState
from swh.objstorage.backends.winery.stats import Stats
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import ObjStorageInterface
from swh.objstorage.objstorage import compute_hash

logger = logging.getLogger(__name__)

WorkerKind = Literal["ro", "rw", "pack", "rbd", "rw_shard_cleaner", "stats"]


def work(
    kind: WorkerKind,
    storage: Union[ObjStorageInterface, Dict[str, Any]],
    time_remaining: datetime.timedelta,
    worker_args: Optional[Dict[WorkerKind, Any]] = None,
    worker_id: int = 0,
) -> WorkerKind:
    if not worker_args:
        worker_args = {}

    kind_args = worker_args.get(kind, {})

    process_name = f"Worker-{kind}-{worker_id}"
    process = current_process()
    if process and process.name != "MainProcess":
        process.name = process_name

    logger.info("Started process %s", process_name)

    application_name = f"Winery Benchmark {process_name}"

    if kind == "ro":
        try:
            if isinstance(storage, dict):
                storage = get_objstorage(
                    cls="winery",
                    application_name=application_name,
                    **{
                        **storage,
                        "readonly": True,
                    },
                )
            return ROWorker(storage, **kind_args).run(time_remaining=time_remaining)
        finally:
            if isinstance(storage, WineryObjStorage):
                storage.on_shutdown()
    elif kind == "rw":
        try:
            if isinstance(storage, dict):
                storage = get_objstorage(
                    cls="winery", application_name=application_name, **storage
                )
            return RWWorker(storage, **kind_args).run(time_remaining=time_remaining)
        finally:
            if isinstance(storage, WineryObjStorage):
                storage.on_shutdown()
    elif kind == "pack":
        return PackWorker(application_name=application_name, **kind_args).run()
    elif kind == "rbd":
        return RBDWorker(application_name=application_name, **kind_args).run()
    elif kind == "rw_shard_cleaner":
        return RWShardCleanerWorker(
            application_name=application_name, **kind_args
        ).run()
    elif kind == "stats":
        return StatsPrinter(application_name=application_name, **kind_args).run(
            time_remaining=time_remaining
        )
    else:
        raise ValueError("Unknown worker kind: %s" % kind)


class Worker:
    def __init__(self, storage: ObjStorageInterface):
        assert isinstance(
            storage, WineryObjStorage
        ), f"winery_benchmark passed unexpected {storage.__class__.__name__}"
        self.stats = Stats(storage.winery.args.get("output_dir"))
        self.storage = storage

    def run(self, time_remaining: datetime.timedelta) -> WorkerKind:
        raise NotImplementedError


class PackWorker:
    def __init__(
        self,
        base_dsn: str,
        shard_max_size: int,
        throttle_read: int,
        throttle_write: int,
        application_name: Optional[str] = None,
        rbd_create_images: bool = True,
        rbd_pool_name: str = "shards",
        output_dir: Optional[str] = None,
    ):
        self.base_dsn = base_dsn
        self.shard_max_size = shard_max_size
        self.output_dir = output_dir
        self.throttle_read = throttle_read
        self.throttle_write = throttle_write
        self.rbd_create_images = rbd_create_images
        self.rbd_pool_name = rbd_pool_name
        self.application_name = application_name
        self.waited = 0

    def stop_packing(self, shards_count: int) -> bool:
        return shards_count >= 1 or self.waited > 60

    def wait_for_shard(self, attempt: int) -> None:
        if self.waited > 60:
            raise ValueError("Shard waited for too long")
        time.sleep(0.1)
        self.waited += 1

    def run(self) -> Literal["pack"]:
        shard_packer(
            base_dsn=self.base_dsn,
            shard_max_size=self.shard_max_size,
            throttle_read=self.throttle_read,
            throttle_write=self.throttle_write,
            rbd_pool_name=self.rbd_pool_name,
            rbd_create_images=self.rbd_create_images,
            rbd_wait_for_image=self.wait_for_shard,
            output_dir=self.output_dir,
            stop_packing=self.stop_packing,
            wait_for_shard=self.wait_for_shard,
            application_name=self.application_name,
        )
        return "pack"


class RBDWorker:
    def __init__(
        self,
        base_dsn: str,
        rbd_pool_name: str,
        rbd_map_options: str,
        shard_max_size: int,
        application_name: Optional[str] = None,
        duration: int = 10,
    ):
        self.base_dsn = base_dsn
        self.pool = Pool(
            shard_max_size=shard_max_size,
            rbd_pool_name=rbd_pool_name,
            rbd_map_options=rbd_map_options,
        )
        self.duration = duration
        self.started = time.monotonic()
        self.application_name = application_name
        self.waited = 0

    def wait_for_shard(self, attempt: int) -> None:
        time.sleep(1)
        self.waited += 1

    def stop_running(self) -> bool:
        return time.monotonic() > self.started + self.duration or self.waited > 5

    def run(self) -> Literal["rbd"]:
        self.pool.manage_images(
            base_dsn=self.base_dsn,
            manage_rw_images=True,
            wait_for_image=self.wait_for_shard,
            stop_running=self.stop_running,
            application_name=self.application_name,
        )
        return "rbd"


class RWShardCleanerWorker:
    def __init__(
        self,
        base_dsn: str,
        min_mapped_hosts: int = 1,
        application_name: Optional[str] = None,
        duration: int = 10,
    ):
        self.base_dsn = base_dsn
        self.min_mapped_hosts = min_mapped_hosts
        self.application_name = application_name
        self.duration = duration
        self.started = time.monotonic()
        self.waited = 0

    def stop_cleaning(self, num_cleaned: int) -> bool:
        return num_cleaned >= 1 or self.waited > 5

    def wait_for_shard(self, attempt: int) -> None:
        time.sleep(1)
        self.waited += 1

    def run(self) -> Literal["rw_shard_cleaner"]:
        rw_shard_cleaner(
            base_dsn=self.base_dsn,
            min_mapped_hosts=self.min_mapped_hosts,
            stop_cleaning=self.stop_cleaning,
            wait_for_shard=self.wait_for_shard,
            application_name=self.application_name,
        )
        return "rw_shard_cleaner"


class StatsPrinter:
    def __init__(
        self,
        base_dsn: str,
        shard_max_size: int,
        application_name: Optional[str] = None,
        interval: int = 5 * 60,
    ):
        self.base_dsn = base_dsn
        self.shard_max_size = shard_max_size
        self.interval = datetime.timedelta(seconds=interval)
        self.application_name = application_name or "Winery Benchmark Stats Printer"
        self.objects_per_shard: Dict[str, int] = {}

    def get_winery_reader(self) -> WineryReader:
        return WineryReader(
            base_dsn=self.base_dsn,
            shard_max_size=self.shard_max_size,
            application_name=self.application_name,
        )

    def run(self, time_remaining: datetime.timedelta) -> Literal["stats"]:
        try:
            return self._run(time_remaining)
        except Exception:
            logger.exception("StatsPrinter.run raised exception")
            return "stats"

    def _run(self, time_remaining: datetime.timedelta) -> Literal["stats"]:
        sleep = min(time_remaining, self.interval).total_seconds()
        if sleep > 1:
            time.sleep(sleep)

        winery = self.get_winery_reader()
        shards = list(winery.base.list_shards())
        shard_counts: Counter[ShardState] = Counter()

        printed_rw_header = False

        for shard_name, _ in shards:
            # Get a fresh version of the state again to try and avoid a race
            state = winery.base.get_shard_state(shard_name)
            shard_counts[state] += 1
            if state not in {ShardState.STANDBY, ShardState.WRITING}:
                if shard_name not in self.objects_per_shard:
                    self.objects_per_shard[shard_name] = winery.base.count_objects(
                        shard_name
                    )
            else:
                if not printed_rw_header:
                    logger.info("read-write shard stats:")
                    printed_rw_header = True

                objects = winery.base.count_objects(shard_name)
                try:
                    shard = winery.rwshard(shard_name)
                    size = shard.size
                except psycopg_pool.PoolTimeout:
                    logger.info(
                        "Shard %s got eaten by the rw shard cleaner, sorry", shard_name
                    )
                    size = 0
                logger.info(
                    " shard %s (state: %s): objects: %s, total_size: %.1f GiB (%2.1f%%)",
                    shard_name,
                    state.name,
                    objects,
                    size / (1024 * 1024 * 1024),
                    100 * size / self.shard_max_size,
                )

        logger.info(
            "Read-only shard stats: count: %s, objects: %s, total_size (est.): %.1f GiB",
            len(self.objects_per_shard),
            sum(self.objects_per_shard.values()),
            (len(self.objects_per_shard) * self.shard_max_size) / (1024 * 1024 * 1024),
        )
        logger.info(
            "Shard counts: %s",
            ", ".join(f"{state.name}: {shard_counts[state]}" for state in ShardState),
        )

        return "stats"


class ROWorker(Worker):
    def __init__(self, storage: ObjStorageInterface, max_request: int = 1000) -> None:
        super().__init__(storage)

        if not isinstance(self.storage.winery, WineryReader):
            raise ValueError(
                f"Running ro benchmark on {self.storage.winery.__class__.__name__}"
                ", expected read-only"
            )

        self.winery: WineryReader = self.storage.winery
        self.max_request = max_request

    def run(self, time_remaining: datetime.timedelta) -> Literal["ro"]:
        try:
            self._ro(time_remaining)
        except psycopg.OperationalError:
            # It may happen when the database is dropped, just
            # conclude the read loop gracefully and move on
            logger.exception("RO worker got exception...")
        finally:
            self.finalize()

        return "ro"

    def _ro(self, time_remaining: datetime.timedelta):
        cutoff = time.time() + time_remaining.total_seconds()
        remaining = self.max_request

        start = time.monotonic()
        tablesample = 0.1
        random_cutoff = 0.1
        while remaining:
            if time.time() > cutoff:
                break
            with self.storage.winery.base.pool.connection() as db:
                limit = min(remaining, 1000)
                c = db.execute(
                    """
                    WITH selected AS (
                      SELECT signature, random() r
                      FROM signature2shard TABLESAMPLE BERNOULLI (%s)
                      WHERE state = 'present' and random() < %s
                      LIMIT %s)
                    SELECT signature FROM selected ORDER BY r
                    """,
                    (
                        tablesample,
                        random_cutoff,
                        limit,
                    ),
                )

                if c.rowcount == 0:
                    logger.info(
                        "Worker(ro, %s): empty (tablesample=%s, random_cutoff=%s), sleeping",
                        os.getpid(),
                        tablesample,
                        random_cutoff,
                    )
                    tablesample = min(tablesample * 10, 100)
                    random_cutoff = min(random_cutoff * 3, 1)
                    time.sleep(1)
                    continue
                elif c.rowcount == limit:
                    tablesample = max(tablesample / 10, 0.1)
                    random_cutoff = max(random_cutoff / 3, 0.1)

                for (obj_id,) in c:
                    remaining -= 1
                    if time.time() > cutoff:
                        remaining = 0
                        break
                    content = self.storage.get(obj_id={"sha256": obj_id})
                    assert content is not None
                    if self.stats.stats_active:
                        self.stats.stats_read(obj_id, content)

        elapsed = time.monotonic() - start
        logger.info("Worker(ro, %s): finished (%.2fs)", os.getpid(), elapsed)

    def finalize(self):
        self.storage.on_shutdown()


class RWWorker(Worker):
    """A read-write benchmark worker

    Args:
      storage: the read-write storage used
      object_limit: the number of objects written before stopping
      single_shard: stop when the worker switches to a new shard
      block_until_packed: whether to wait for shards to be packed before exiting
    """

    def __init__(
        self,
        storage: ObjStorageInterface,
        object_limit: Optional[int] = None,
        single_shard: bool = True,
        block_until_packed: bool = True,
    ) -> None:
        super().__init__(storage)

        if not isinstance(self.storage.winery, WineryWriter):
            raise ValueError(
                f"Running rw benchmark on {self.storage.winery.__class__.__name__}"
                ", expected read-write"
            )

        self.winery: WineryWriter = self.storage.winery
        self.object_limit = object_limit
        self.single_shard = single_shard
        self.block_until_packed = block_until_packed
        self.count = 0

    def payloads_define(self):
        self.payloads = [
            3 * 1024 + 1,
            3 * 1024 + 1,
            3 * 1024 + 1,
            3 * 1024 + 1,
            3 * 1024 + 1,
            10 * 1024 + 1,
            13 * 1024 + 1,
            16 * 1024 + 1,
            70 * 1024 + 1,
            80 * 1024 + 1,
        ]

    def run(self, time_remaining: datetime.timedelta) -> Literal["rw"]:
        end = time.monotonic() + time_remaining.total_seconds()
        self.payloads_define()
        random_content = open("/dev/urandom", "rb")
        logger.info("Worker(rw, %s): start", os.getpid())
        start = time.monotonic()
        while self.keep_going() and time.monotonic() < end:
            content = random_content.read(random.choice(self.payloads))
            obj_id = compute_hash(content, "sha256")
            self.storage.add(content=content, obj_id=obj_id)
            if self.stats.stats_active:
                self.stats.stats_write(obj_id, content)
            self.count += 1
        self.finalize()
        elapsed = time.monotonic() - start
        logger.info("Worker(rw, %s): finished (%.2fs)", os.getpid(), elapsed)

        return "rw"

    def keep_going(self) -> bool:
        if self.object_limit is not None and self.count >= self.object_limit:
            return False
        if self.single_shard and self.winery.shards_filled:
            return False

        return True

    def finalize(self):
        self.storage.on_shutdown()

        if not self.block_until_packed:
            return
        logger.info(
            "Worker(rw, %s): waiting for %s objects to be packed",
            os.getpid(),
            self.count,
        )
        for packer in self.winery.packers:
            packer.join()
            assert packer.exitcode == 0


class Bench(object):
    def __init__(
        self,
        storage_config: Union[ObjStorageInterface, Dict[str, Any]],
        duration: int,
        workers_per_kind: Dict[WorkerKind, int],
        worker_args: Optional[Dict[WorkerKind, Any]] = None,
    ) -> None:
        self.storage_config = storage_config
        self.duration = duration
        self.workers_per_kind = workers_per_kind
        self.worker_args = worker_args or {}
        self.start = 0

    def timer_start(self):
        self.start = time.monotonic()

    def timeout(self) -> bool:
        return time.monotonic() - self.start > self.duration

    def time_remaining(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self.start + self.duration - time.monotonic())

    async def run(self) -> int:
        self.timer_start()

        loop = asyncio.get_running_loop()

        workers_count = sum(self.workers_per_kind.values())

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=workers_count
        ) as executor:
            logger.info("Running winery benchmark")

            self.count = 0
            workers: "Set[asyncio.Future[WorkerKind]]" = set()

            def create_worker(kind: WorkerKind) -> "asyncio.Future[WorkerKind]":
                self.count += 1
                logger.info("launched %s worker number %s", kind, self.count)
                return loop.run_in_executor(
                    executor,
                    work,
                    kind,
                    self.storage_config,
                    self.time_remaining(),
                    self.worker_args,
                    self.count,
                )

            for kind, count in self.workers_per_kind.items():
                for _ in range(count):
                    workers.add(create_worker(kind))

            while len(workers) > 0:
                logger.info(
                    "Waiting for %s workers",
                    ", ".join(
                        f"{v} {k}" for k, v in self.workers_per_kind.items() if v
                    ),
                )
                current = workers
                done, pending = await asyncio.wait(
                    current, return_when=asyncio.FIRST_COMPLETED
                )
                workers = pending
                exceptions = list(filter(None, [task.exception() for task in done]))
                if exceptions:
                    for task in pending:
                        task.cancel()
                    if sys.version_info >= (3, 11):
                        raise BaseExceptionGroup(  # noqa: F821
                            "Some workers raised an exception", exceptions
                        )
                    else:
                        for exc in exceptions:
                            logger.error("Worker raised an exception", exc_info=exc)
                        raise exceptions[0]

                for task in done:
                    kind = task.result()
                    logger.info("worker %s complete", kind)
                    if not self.timeout():
                        workers.add(create_worker(kind))
                    else:
                        self.workers_per_kind[kind] -= 1

            logger.info("Bench.run: finished")

        return self.count



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict

from typing_extensions import Literal

ID_HASH_ALGO: Literal["sha1"] = "sha1"

ID_HEXDIGEST_LENGTH = 40
"""Size in bytes of the hash hexadecimal representation."""

ID_DIGEST_LENGTH = 20
"""Size in bytes of the hash"""

DEFAULT_LIMIT = 10000
"""Default number of results of ``list_content``."""

VALID_HEXCHARS = frozenset("0123456789abcdef")
"""Valid characters for hexadecimal values"""

ID_HEXDIGEST_LENGTH_BY_ALGO: Dict[Literal["sha1", "sha256"], int] = {
    "sha1": 40,
    "sha256": 64,
}
"""Length of a valid hexdigest for each "primary" algorithm"""


def is_valid_hexdigest(hexdigest: str, algo: Literal["sha1", "sha256"]):
    """Return whether `hexdigest` is a valid hexdigest for the given `algo`."""
    return (
        len(hexdigest) == ID_HEXDIGEST_LENGTH_BY_ALGO[algo]
        and set(hexdigest) <= VALID_HEXCHARS
    )



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Iterator

from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import ObjStorage, timed


class InMemoryObjStorage(ObjStorage):
    """In-Memory objstorage.

    Intended for test purposes.

    """

    PRIMARY_HASH = "sha1"
    name: str = "memory"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.state: Dict[bytes, bytes] = {}

    def check_config(self, *, check_write):
        return True

    def _state_key(self, obj_id: ObjId) -> bytes:
        if isinstance(obj_id, dict):
            return obj_id[self.PRIMARY_HASH]
        else:
            return obj_id

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        return self._state_key(obj_id) in self.state

    def __iter__(self) -> Iterator[CompositeObjId]:
        for id_ in sorted(self.state):
            yield {self.PRIMARY_HASH: id_}

    @timed
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        if check_presence and obj_id in self:
            return

        self.state[self._state_key(obj_id)] = content

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        if obj_id not in self:
            raise ObjNotFoundError(obj_id)

        return self.state[self._state_key(obj_id)]

    def delete(self, obj_id: ObjId):
        super().delete(obj_id)  # Check delete permission
        if obj_id not in self:
            raise ObjNotFoundError(obj_id)

        self.state.pop(self._state_key(obj_id))
        return True



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
from collections import OrderedDict
from datetime import timedelta
from io import BytesIO
from typing import Iterator, Optional
from urllib.parse import urlencode

from libcloud.storage import providers
import libcloud.storage.drivers.s3
from libcloud.storage.types import ObjectDoesNotExistError, Provider

from swh.model import hashutil
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import (
    CompressionFormat,
    ObjStorage,
    objid_to_default_hex,
    timed,
)


def patch_libcloud_s3_urlencode():
    """Patches libcloud's S3 backend to properly sign queries.

    Recent versions of libcloud are not affected (they use signature V4),
    but 1.5.0 (the one in Debian 9) is."""

    def s3_urlencode(params):
        """Like urllib.parse.urlencode, but sorts the parameters first.
        This is required to properly compute the request signature, see
        https://docs.aws.amazon.com/AmazonS3/latest/dev/RESTAuthentication.html#ConstructingTheCanonicalizedResourceElement
        """  # noqa
        return urlencode(OrderedDict(sorted(params.items())))

    libcloud.storage.drivers.s3.urlencode = s3_urlencode


patch_libcloud_s3_urlencode()


class CloudObjStorage(ObjStorage, metaclass=abc.ABCMeta):
    """Abstract ObjStorage that connect to a cloud using Libcloud

    Implementations of this class must redefine the _get_provider
    method to make it return a driver provider (i.e. object that
    supports `get_driver` method) which return a LibCloud driver (see
    https://libcloud.readthedocs.io/en/latest/storage/api.html).

    Args:
      container_name: Name of the base container
      path_prefix: prefix to prepend to object paths in the container,
                   separated with a slash
      compression: compression algorithm to use for objects
      kwargs: extra arguments are passed through to the LibCloud driver
    """

    PRIMARY_HASH = "sha1"
    name: str = "cloud"

    def __init__(
        self,
        container_name: str,
        compression: CompressionFormat = "gzip",
        path_prefix: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.driver = self._get_driver(**kwargs)
        self.container_name = container_name
        self.container = self.driver.get_container(container_name=container_name)
        self.compression = compression
        self.path_prefix = None
        if path_prefix:
            self.path_prefix = path_prefix.rstrip("/") + "/"

    def _get_driver(self, **kwargs):
        """Initialize a driver to communicate with the cloud

        Kwargs: arguments passed to the StorageDriver class, typically
          key: key to connect to the API.
          secret: secret key for authentication.
          secure: (bool) support HTTPS
          host: (str)
          port: (int)
          api_version: (str)
          region: (str)

        Returns:
            a Libcloud driver to a cloud storage.

        """
        # Get the driver class from its description.
        cls = providers.get_driver(self._get_provider())
        # Initialize the driver.
        return cls(**kwargs)

    @abc.abstractmethod
    def _get_provider(self):
        """Get a libcloud driver provider

        This method must be overridden by subclasses to specify which
        of the native libcloud driver the current storage should
        connect to.  Alternatively, provider for a custom driver may
        be returned, in which case the provider will have to support
        `get_driver` method.

        """
        raise NotImplementedError(
            "%s must implement `get_provider` method" % type(self)
        )

    def check_config(self, *, check_write):
        """Check the configuration for this object storage"""
        # FIXME: hopefully this blew up during instantiation
        return True

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        try:
            self._get_object(obj_id)
        except ObjNotFoundError:
            return False
        else:
            return True

    def __iter__(self) -> Iterator[CompositeObjId]:
        """Iterate over the objects present in the storage

        Warning: Iteration over the contents of a cloud-based object storage
        may have bad efficiency: due to the very high amount of objects in it
        and the fact that it is remote, get all the contents of the current
        object storage may result in a lot of network requests.

        You almost certainly don't want to use this method in production.
        """
        for obj in self.driver.iterate_container_objects(self.container):
            name = obj.name

            if self.path_prefix and not name.startswith(self.path_prefix):
                continue

            if self.path_prefix:
                name = name[len(self.path_prefix) :]

            yield {self.PRIMARY_HASH: hashutil.hash_to_bytes(name)}

    def __len__(self):
        """Compute the number of objects in the current object storage.

        Warning: this currently uses `__iter__`, its warning about bad
        performance applies.

        Returns:
            number of objects contained in the storage.

        """
        return sum(1 for i in self)

    @timed
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        if check_presence and obj_id in self:
            return

        self._put_object(content, obj_id)

    def restore(self, content: bytes, obj_id: ObjId) -> None:
        return self.add(content, obj_id, check_presence=False)

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        obj = b"".join(self._get_object(obj_id).as_stream())
        return self.decompress(obj, objid_to_default_hex(obj_id))

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        return self._get_object(obj_id).get_cdn_url()

    def delete(self, obj_id: ObjId):
        super().delete(obj_id)  # Check delete permission
        obj = self._get_object(obj_id)
        return self.driver.delete_object(obj)

    def _object_path(self, obj_id: ObjId) -> str:
        """Get the full path to an object"""
        if isinstance(obj_id, dict):
            obj_id = obj_id[self.PRIMARY_HASH]

        hex_obj_id = hashutil.hash_to_hex(obj_id)
        if self.path_prefix:
            return self.path_prefix + hex_obj_id
        else:
            return hex_obj_id

    def _get_object(self, obj_id: ObjId):
        """Get a Libcloud wrapper for an object pointer.

        This wrapper does not retrieve the content of the object
        directly.

        """
        object_path = self._object_path(obj_id)

        try:
            return self.driver.get_object(self.container_name, object_path)
        except ObjectDoesNotExistError:
            raise ObjNotFoundError(obj_id)

    def _put_object(self, content, obj_id):
        """Create an object in the cloud storage.

        Created object will contain the content and be referenced by
        the given id.

        """
        object_path = self._object_path(obj_id)
        self.driver.upload_object_via_stream(
            BytesIO(self.compress(content)),
            self.container,
            object_path,
        )


class AwsCloudObjStorage(CloudObjStorage):
    """Amazon's S3 Cloud-based object storage"""

    name: str = "s3"

    def _get_provider(self):
        return Provider.S3

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        return self.driver.get_object_cdn_url(
            self._get_object(obj_id),
            ex_expiry=(expiry.total_seconds() / 3600) if expiry is not None else 24,
        )


class OpenStackCloudObjStorage(CloudObjStorage):
    """OpenStack Swift Cloud based object storage"""

    name: str = "swift"

    def _get_provider(self):
        return Provider.OPENSTACK_SWIFT
